/*
**   IBM Omni driver
**   Copyright (c) International Business Machines Corp., 2002
**
**   This library is free software; you can redistribute it and/or modify
**   it under the terms of the GNU Lesser General Public License as published
**   by the Free Software Foundation; either version 2.1 of the License, or
**   (at your option) any later version.
**
**   This library is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
**   the GNU Lesser General Public License for more details.
**
**   You should have received a copy of the GNU Lesser General Public License
**   along with this library; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
**
** Idea borrowed from http://hpinkjet.sourceforge.net
**
** Written by Mark Hamzy hamzy@us.ibm.com
**
** TODO
**    - send more than 1 scanline at a time
**    - only support error diffusion
**    - add all forms
**    - make sure form names are acceptable
**    - make sure media names are acceptable
**    - add color correction to job properties
**    - add brightness, contrast, and saturation to job properties
**
** Supported printers:
**    Stylus Color 980
**    Stylus Color 880
**    Stylus Color 777
**    Stylus Color 680
**    Stylus Color 580
**    Stylus Color 480
**    Stylus C70
**    Stylus C60
**    Stylus C40
**    Stylus C20
**    Stylus Photo 810
**    Stylus Photo 820
**
** default job properties for sc60:
**    form=FORM_LETTER
**    tray=TRAY_AUTO
**    media=MEDIA_PLAIN
**    resolution=RESOLUTION_720_X_720
**    printmode=PRINT_MODE_24_CMYK
**    quality=super-fine
**    high-speed=off
**    color-mode=none
**
** form=
** -----
** FORM_A4
** FORM_B5
** FORM_A5
** FORM_LETTER
** FORM_LEGAL
** FORM_EXECUTIVE
** FORM_HALF_LETTER
** FORM_A6
** FORM_5_X_8_CARD
** FORM_8_X_10_CARD
** FORM_DL_ENVELOPE
** FORM_C6_ENVELOPE
** FORM_4_X_6_CARD
** FORM_PHOTO_100_150
** FORM_5_X_7
** FORM_PANORAMIC
**
** media=                   tray=     resolution=          quality=    high-speed=    color-mode=
** ------                   -----     -----------          --------    -----------    -----------
** MEDIA_PLAIN              TRAY_AUTO RESOLUTION_120_X_120 draft       on          ED auto/photo/graph      320*120
**                                    RESOLUTION_360_X_360 standard    on/off      ED auto/photo/graph      360*360
**                                    RESOLUTION_360_X_360 fine        on/off      ED auto/photo/graph      360*720
**                                    RESOLUTION_360_X_360 super-fine  on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 super-fine  on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 photo       on/off      ED auto/photo/graph      1440*720
**
** MEDIA_SPECIAL_360        TRAY_AUTO RESOLUTION_360_X_360 standard    off         ED auto/photo/graph      360*360
**
** MEDIA_HIGH_GLOSS_FILM    TRAY_AUTO RESOLUTION_360_X_360 standard    off         ED auto/photo/graph      360*720
**                                    RESOLUTION_720_X_720 standard    off         ED auto/photo/graph      360*720
**                                    RESOLUTION_360_X_360 fine        on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 fine        on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 photo       on/off      ED auto/photo/graph      1440*720
**
** MEDIA_PHOTOGRAPHIC_PAPER TRAY_AUTO RESOLUTION_360_X_360 standard    off         ED auto/photo/graph      360*720
**                                    RESOLUTION_720_X_720 standard    off         ED auto/photo/graph      360*720
**                                    RESOLUTION_360_X_360 fine        on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 fine        on/off      ED auto/photo/graph      720*720
**                                    RESOLUTION_720_X_720 photo       on/off      ED auto/photo/graph      1440*720
**                                    RESOLUTION_720_X_720 super-photo off         ED auto/photo/graph      2880*720
**
** MEDIA_HIGH_GLOSS_FILM    TRAY_AUTO RESOLUTION_720_X_720 photo       on/off      ED auto/photo/graph      1440*720
**                                    RESOLUTION_720_X_720 super-photo off         ED auto/photo/graph      2880*720
**
** MEDIA_TRANSPARENCY       TRAY_AUTO RESOLUTION_360_X_360 standard    off         ED auto/photo/graph      360*360
**
** Example usage:
**
** gs -dNOPAUSE -dBATCH -sDEVICE=omni -sDeviceName=Epson_Stylus_C60 -sproperties="resolution=RESOLUTION_360_X_360 quality=fine" -sOutputFile=/tmp/file.out file.ps
*/
#include "PDCBlitterClient.hpp"
#include "OmniServer.hpp"
#include "PrinterCommand.hpp"
#include "Omni.hpp"
#include "JobProperties.hpp"

/****************************************************************************/
/* Epson changable...                                                       */
/****************************************************************************/

/****************************************************************************/
/* Epson headers                                                            */
/****************************************************************************/

typedef struct _POINT {
   long x;
   long y;
} POINT, *PPOINT;

typedef struct _HCC {
   int cx;
   int cy;
   int iLeft;
   int iTop;
   int iRight;
   int iBottom;
} HCC, *PHCC;

typedef struct _LIBPRT_INIT_STRUCT {
   FILE* output;         // The pointer to output.
   PSZRO paper;          // Paper
                         //  "A4"        :A4  210x297 mm
                         //  "B5"        :B5  182x257 mm
                         //  "A5"        :A5 148x210 mm
                         //  "LT"        :Letter  8 1/2x11 in
                         //  "LGL"       :Legal  8 1/2x14 in
                         //  "EXE"       :Executive  7 1/4x10 1/2 in
                         //  "HLT"       :Half Letter  5 1/2x8 1/2 in
                         //  "A6"        :A6  105x148 mm
                         //  "INDEX5"    :Index card  5x8 in
                         //  "INDEX8"    :Index card  8x10 in
                         //  "ENV10"     :Envelope #10  4 1/8x9 1/2
                         //  "ENVDL"     :Envelope DL  110x220 mm
                         //  "ENVC6"     :Envelope C6  114x162 mm
                         //  "ENV5X8"    :Envelope 132x220 mm
                         //  "PHOTO"     :Photo Paper  4x6 in
                         //  "4X6FULL"   :Photo Paper 4 x 6 in No Perforation
                         //  "PP100"     :Photo Paper 100x150 mm
                         //  "PP200"     :Photo Paper 200x300 mm
                         //  "L"         :3.5 x 5 in  89 x 127 mm
                         //  "2L"        :5 x 7 in  127 x 178 mm
                         //  "PANORAM"   :Panoramic 210x594 mm

   PSZRO format;         // Print format
                         //  "MONO"  :Black print
                         //  "COLOR" :Color print

   PSZRO resol;          // Resolution
                         //  "120"   :120dpi
                         //  "360"   :360dpi
                         //  "720"   :720dpi

   PSZRO high_speed;     // High Speed Print
                         //  "ON"    :Bi-directional Print
                         //  "OFF"   :Simplex Print

   PSZRO bin_id;         // Paper Source
                         //  "AUTO"    :Auto Sheet Feeder

   PSZRO media_type;     // Media Type
                         //  "PLAIN"  :Plain Paper
                         //  "FINE"   :360dpi Ink Jet Paper
                         //  "SFINE"  :Photo Quality Ink Jet Paper
                         //  "GPAPER" :Photo Paper
                         //  "GFILM"  :Photo Quality Glossy Film
                         //  "OHP"    :Ink Jet Transparencies

   PSZRO qlevel;         // Quality Level
                         //  "DRAFT"  :Draft Print
                         //  "STD"    :Standard Print
                         //  "FINE"   :Fine Print
                         //  "SFINE"  :Super Fine Print
                         //  "PHOTO"  :High Print
                         //  "SPHOTO" :Super High Print

   POINT margin;         // Margin
                         //  x = Top Margin[0 - bottom(maximum) value](pixel)
                         //  y = Left Margin[0 - right(maximum) value](pixel)

   POINT paper_size;     // Paper Size
                         //  x = Width(pixel)
                         //  y = Length(pixel)

   POINT paper_area;     // Print Area
                         //  x = Width of printable area(pixel)
                         //  y = Length of printable area(pixel)

   PSZRO mw_type;        // Micro Weave
                         //  "ON"    :MicroWeave ON
                         //  "OFF"   :MicroWeave OFF
/// was:
///void* reserved;       // reserved
///                      //  blank
} LIBPRT_INIT, *LP_LIBPRT_INIT;

/*
** Combination of options
**
** The list shown below indicates the combination of the option that is available
** to the Photo Image Print System for Stylus Color 580.  Refer to the list when
** changing the options from the command line.
**
** Black & Color print
**
**      Media        Resolution  Quality   High   Halftone        Color          Output
**                                        Speed                    mode         resolution
**                                                         (Color print only)   (reference)
**
**         -mt         -re         -ql     -hs     -ht      -cc
**         PLAIN       120         DRAFT   ON      ED       AUTO/PHOTO/GRAPH    320*120
**                     360         STD     ON/OFF  ED       AUTO/PHOTO/GRAPH    360*360
**                     360         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    360*720
**                     360         SFINE   ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         SFINE   ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**
**         FINE        360         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*360
**
**         SFINE       360         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     720         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     360         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**
**         PMMATT      360         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     720         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     360         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**
**         GPAPER      360         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     720         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*720
**                     360         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         FINE    ON/OFF  ED       AUTO/PHOTO/GRAPH    720*720
**                     720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**                     720         SPHOTO  OFF     ED       AUTO/PHOTO/GRAPH    2880*720
**
**         PMPHOTO     720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**                     720         SPHOTO  OFF     ED       AUTO/PHOTO/GRAPH    2880*720
**
**         GFILM       720         PHOTO   ON/OFF  ED       AUTO/PHOTO/GRAPH    1440*720
**                     720         SPHOTO  OFF     ED       AUTO/PHOTO/GRAPH    2880*720
**
**         OHP         360         STD     OFF     ED       AUTO/PHOTO/GRAPH    360*360
**
** The size of the each paper and the margin is listed below.
**
** [Auto Sheet Feeder]
** PaperSize                       Paper Size(1mm)  Margin(1mm)
** Name                            Width  Height    Left     Top   Right   Bottom
** A4 210 x 297 mm                 210    297       3        3     3       14
** B5 182 x 257 mm                 182    257       3        3     3       14
** A5 148 x 210 mm                 148    210       3        3     3       14
** Letter 8 1/2 x 11 in            216    279       3        3     3       14
** Legal 8 1/2 x 14 in             216    356       3        3     3       14
** Executive 7 1/4 x 10 1/2 in     184    267       3        3     3       14
** Half Letter 5 1/2 x 8 1/2 in    140    216       3        3     3       14
** A6 Index card 105 x 148 mm      105    148       3        3     3       14
** Index card 5 x 8 in             127    203       3        3     3       14
** Index card 8 x 10 in            203    254       3        3     3       14
** Envelope #10 4 1/8 x 9 1/2 in   241    105       3        3     28      14
** Envelope DL 110 x 220 mm        220    110       3        3     7       14
** Envelope C6 114 x 162 mm        162    114       3        3     3       14
** Envelope 132 x 220 mm           220    132       3        3     7       14
** Photo Paper 4 x 6 in            114    164       3        3     3       3
** Photo Paper 100 x 150 mm        100    150       3        3     3       14
** Photo Paper 200 x 300 mm        216    338       3        14    3       14
** 4 x 6 in No Perforations        102    152       3        3     3       14
** 3.5 x 5 in  89 x 127 mm          89    127       3        3     3       14
** 5 x 7 in  127 x 178 mm          127    178       3        3     3       14
** Panoramic 210 x 594 mm          210    594       3        3     3       14
*/

typedef struct _LIBPRT_PAGE_STRUCT {
   PSZRO color_mode;     // Color mode adjustment
                         //  "AUTO"  :Automatic setting
                         //  "PHOTO" :Photo - realistic
                         //  "GRAPH" :Vivid color
                         //  "NONE"  :No adjustment

   PSZRO halftone_type;  // Halftone Type
                         //  "ED"    :High Quality Halftoning

   int brightness;       // Brightness                 = [-50 - 50]
   int contrast;         // Contrast                   = [-50 - 50]
   int saturation;       // Saturation                 = [-50 - 50]
   int r_strength;       // Red color compensation     = [-50 - 50]
   int g_strength;       // Green color compensation   = [-50 - 50]
   int b_strength;       // Blue color compensation    = [-50 - 50]
   void* reserved;       // reserved
} LIBPRT_PAGE, *LP_LIBPRT_PAGE;

typedef struct _LIBPRT_ATBMP_BAND_STRUCT {
   long width_bytes;                   // Values of the raster's byte
   unsigned char* byte_data;           // Attribute Bitmap of raster
} LIBPRT_ATBMP_BAND, *LP_LIBPRT_ATBMP_BAND;

typedef struct _LIBPRT_BAND_STRUCT {
   long width_pixel;                   // Values of the raster's pixel
   long width_bytes;                   // Values of the raster's byte
   long line_number;                   // Values of the raster's number
   unsigned char* byte_data;           // Image of raster
   LP_LIBPRT_ATBMP_BAND lp_atbmp_band; // "Color mode adjustment" sets LP_LIBPRT_ATBMP_BAND in case of "AUTO".
                                       // Set null besides it.
} LIBPRT_BAND, *LP_LIBPRT_BAND;

typedef int (*PFNLIBPRT_INIT)      (HANDLE*,            // The pointer to HANDLE used in the internal library.
                                    LP_LIBPRT_INIT);    // Initialization structure of the library.
typedef int (*PFNLIBPRT_PAGE_INIT) (HANDLE,             // The variable used in the internal library.
                                    LP_LIBPRT_PAGE);    // Initialization structure of the page print.
typedef int (*PFNLIBPRT_OUT)       (HANDLE,             // The variable used in the internal library.
                                    LP_LIBPRT_BAND);    // Initialization structure of the print data.
typedef int (*PFNLIBPRT_OUT_AT)    (HANDLE,             // The variable used in the internal library.
                                    LP_LIBPRT_BAND);    // Initialization structure of the print data.
typedef int (*PFNLIBPRT_PAGE_END)  (HANDLE,             // The variable used in the internal library.
                                    int);               // Cancel flag
                                                        //   When the process is interrupted by some reasons,
                                                        //   it sets the cancel flag to TRUE.
                                                        //   The cancel flag is normally set at FALSE.
typedef int (*PFNLIBPRT_END)       (HANDLE);            // The variable used in the internal library.

typedef enum _State {
   STATE_START,
   STATE_INITED
} STATE, *PSTATE;

typedef enum _DeviceID {
   DEVICEID_NONE                 = 0,
   DEVICEID_STYLUS_COLOR_980,
   DEVICEID_STYLUS_COLOR_880,
   DEVICEID_STYLUS_COLOR_777,
   DEVICEID_STYLUS_COLOR_680,
   DEVICEID_STYLUS_COLOR_580,
   DEVICEID_STYLUS_COLOR_480,
   DEVICEID_STYLUS_C70,
   DEVICEID_STYLUS_C60,
   DEVICEID_STYLUS_C40,
   DEVICEID_STYLUS_C20,
   DEVICEID_STYLUS_PHOTO_810,
   DEVICEID_STYLUS_PHOTO_820
} DEVICEID, *PDEVICEID;

typedef struct _EpsonInstance {
   DEVICEID             eDeviceID;

   HANDLE               hMagicCookie;

   bool                 fShouldCloseOutput;
   FILE                *pfOutput;

   PFNLIBPRT_INIT       pfnLibPrt_Init;
   PFNLIBPRT_PAGE_INIT  pfnLibPrt_Page_Init;
   PFNLIBPRT_OUT        pfnLibPrt_Out;
   PFNLIBPRT_OUT_AT     pfnLibPrt_Out_At;
   PFNLIBPRT_PAGE_END   pfnLibPrt_Page_End;
   PFNLIBPRT_END        pfnLibPrt_End;

   PSZRO                pszQuality;
   PSZRO                pszHighSpeed;
   PSZRO                pszColorMode;
   PSZRO                pszMicroWeave;

   bool                 fInAutoMode;

   bool                 fInMonoMode;

   int                  cbAllocatedBandAt;
   LIBPRT_ATBMP_BAND    bandAt;

   STATE                eState;

   LIBPRT_INIT          init;

} EPSONINSTANCE, *PEPSONINSTANCE;

#define JOBPROP_QUALITY     "quality"
#define   JOBKEY_DRAFT         "draft"
#define   JOBKEY_STANDARD      "standard"
#define   JOBKEY_FINE          "fine"
#define   JOBKEY_SUPER_FINE    "super-fine"
#define   JOBKEY_PHOTO         "photo"
#define   JOBKEY_SUPER_PHOTO   "super-photo"
#define JOBPROP_HIGH_SPEED  "high-speed"
#define   JOBKEY_ON            "on"
#define   JOBKEY_OFF           "off"
#define JOBPROP_COLOR_MODE  "color-mode"
#define   JOBKEY_AUTO          "auto"
#define   JOBKEY_PHOTO         "photo"
#define   JOBKEY_GRAPH         "graph"
#define   JOBKEY_NONE          "none"
#define JOBPROP_MICRO_WEAVE "micro-weave"

static char *vpszDefaultQuality    = JOBKEY_SUPER_FINE;
static char *vpszDefaultHighSpeed  = JOBKEY_ON;
static char *vpszDefaultColorMode  = JOBKEY_PHOTO;
static char *vpszDefaultMicroWeave = JOBKEY_OFF;

static PSZ apszQualities[] = {
   JOBPROP_QUALITY "=" JOBKEY_DRAFT,
   JOBPROP_QUALITY "=" JOBKEY_STANDARD,
   JOBPROP_QUALITY "=" JOBKEY_FINE,
   JOBPROP_QUALITY "=" JOBKEY_SUPER_FINE,
   JOBPROP_QUALITY "=" JOBKEY_PHOTO,
   JOBPROP_QUALITY "=" JOBKEY_SUPER_PHOTO
};

static PSZ apszHighSpeeds[] = {
   JOBPROP_HIGH_SPEED "=" JOBKEY_ON,
   JOBPROP_HIGH_SPEED "=" JOBKEY_OFF
};

static PSZ apszColorModes[] = {
   JOBPROP_COLOR_MODE "=" JOBKEY_AUTO,
   JOBPROP_COLOR_MODE "=" JOBKEY_PHOTO,
   JOBPROP_COLOR_MODE "=" JOBKEY_GRAPH,
   JOBPROP_COLOR_MODE "=" JOBKEY_NONE,
};

static PSZ apszMicroWeaves[] = {
   JOBPROP_MICRO_WEAVE "=" JOBKEY_ON,
   JOBPROP_MICRO_WEAVE "=" JOBKEY_OFF
};

/* Epson helper function prototypes ...
*/
PSZCRO     newCopy                  (PSZCRO          pszString);
PSZCRO     newCopy                  (PSZCRO          pszStart,
                                     PSZCRO          pszEnd);
PSZCRO     getValue                 (PSZCRO          pszEntries,
                                     PSZCRO          pszKey);
PSZCRO     convertToEpsonHighSpeed  (PSZCRO          pszHighSpeedName);
PSZCRO     convertToEpsonQuality    (PSZCRO          pszQualityName);
PSZCRO     convertToEpsonColorMode  (PEPSONINSTANCE  pInstance,
                                     PSZCRO          pszColorMode);
PSZCRO     convertToEpsonMicroWeave (PSZCRO          pszMicroWeaveName);

/****************************************************************************/
/* Epson helper functions                                                   */
/****************************************************************************/

PSZCRO
newCopy (PSZCRO pszString)
{
   PSZ pszReturn = 0;

   if (pszString)
   {
      pszReturn = (PSZ)malloc (strlen (pszString) + 1);

      if (pszReturn)
      {
         strcpy (pszReturn, pszString);
      }
   }

   return pszReturn;
}

PSZCRO
newCopy (PSZCRO pszStart,
         PSZCRO pszEnd)
{
   PSZ pszReturn = 0;

   if (pszStart)
   {
      int cbSize = (int)pszEnd - (int)pszStart + 1;

      pszReturn = (PSZ)malloc (cbSize);

      if (pszReturn)
      {
         strncpy (pszReturn, pszStart, cbSize);
         pszReturn[cbSize - 1] = '\0';
      }
   }

   return pszReturn;
}

PSZCRO
getValue (PSZRO  pszEntries,
          PSZCRO pszKey)
{
   if (!pszEntries)
   {
      return 0;
   }

   PSZRO pszNeedle = 0;

   do
   {
      pszNeedle = strstr (pszEntries, pszKey);

      if (pszNeedle)
      {
         pszEntries = pszNeedle + strlen (pszKey);

         if ('=' == *pszEntries)
         {
            PSZCRO pszStart = pszEntries + 1;
            PSZRO  pszEnd   = pszStart + strlen (pszStart);
            PSZCRO pszSpace = strchr (pszStart, ' ');

            if (pszSpace)
               pszEnd = pszSpace;

            return newCopy (pszStart, pszEnd);
         }
      }
      else
      {
         pszEntries = 0;
      }

   } while (pszEntries);

   return 0;
}

PSZCRO
convertToEpsonHighSpeed (PSZCRO pszHighSpeedName)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": high-speed = " << SAFE_PRINT_PSZ(pszHighSpeedName) << std::endl;
#endif

   PSZRO pszRet = 0;

   if (!pszHighSpeedName)
   {
      return pszRet;
   }

   if (0 == strncmp (pszHighSpeedName, JOBKEY_ON, 2))
   {
      pszRet = "ON";
   }
   else if (0 == strncmp (pszHighSpeedName, JOBKEY_OFF, 3))
   {
      pszRet = "OFF";
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": returning " << SAFE_PRINT_PSZ(pszRet) << std::endl;
#endif

   free ((void *)pszHighSpeedName);

   return pszRet;
}

PSZCRO
convertToEpsonQuality (PSZCRO pszQualityName)
{
#if 0
SC20S
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "PHOTO"    High Print
SC40S
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "PHOTO"    High Print
SC60S
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
SC70S
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
SP810/SP820
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
SC480
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "PHOTO"    High Print
SC580
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "PHOTO"    High Print
SC680/SC777
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
SC880
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
SC980
                    "DRAFT"    Draft Print
                    "STD"      Standard Print
                    "FINE"     Fine Print
                    "SFINE"    Super Fine Print
                    "PHOTO"    High Print
                    "SPHOTO"   Super High Print
#endif
   /*
   ** "DRAFT"    Draft Print
   ** "STD"      Standard Print
   ** "FINE"     Fine Print
   ** "SFINE"    Super Fine Print
   ** "PHOTO"    High Print
   ** "SPHOTO"   Super High Print
   */
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": " << JOBPROP_QUALITY << " = " << SAFE_PRINT_PSZ(pszQualityName) << std::endl;
#endif

   PSZRO pszRet = 0;

   if (!pszQualityName)
   {
      return pszRet;
   }

   if (0 == strncmp (pszQualityName, JOBKEY_DRAFT, 5))
   {
      pszRet = "DRAFT";
   }
   else if (0 == strncmp (pszQualityName, JOBKEY_STANDARD, 8))
   {
      pszRet = "STD";
   }
   else if (0 == strncmp (pszQualityName, JOBKEY_FINE, 4))
   {
      pszRet = "FINE";
   }
   else if (0 == strncmp (pszQualityName, JOBKEY_SUPER_FINE, 10))
   {
      pszRet = "SFINE";
   }
   else if (0 == strncmp (pszQualityName, JOBKEY_PHOTO, 5))
   {
      pszRet = "PHOTO";
   }
   else if (0 == strncmp (pszQualityName, JOBKEY_SUPER_PHOTO, 11))
   {
      pszRet = "SPHOTO";
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": returning " << SAFE_PRINT_PSZ(pszRet) << std::endl;
#endif

   free ((void *)pszQualityName);

   return pszRet;
}

PSZCRO
convertToEpsonColorMode (PEPSONINSTANCE  pInstance,
                         PSZCRO          pszColorModeName)
{
#if 0
SC20S
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC40S
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC60S
                   AUTO     Automatic setting ()
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC70S
                   AUTO     Automatic setting ()
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC810/SC820
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC480
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC580
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC680/SC777
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC880
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
SC980
                   PHOTO    Photo - realistic
                   GRAPH    Vivid color
                   NONE     No adjustment
#endif
   /*
   ** "AUTO"     Automatic setting
   ** "PHOTO"    Photo - realistic
   ** "GRAPH"    Vivid color
   ** "NONE"     No adjustment
   */
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": color-mode = " << SAFE_PRINT_PSZ(pszColorModeName) << std::endl;
#endif

   PSZRO pszRet = 0;

   if (!pszColorModeName)
   {
      return pszRet;
   }

   if (0 == strncmp (pszColorModeName, JOBKEY_PHOTO, 5))
   {
      pszRet = "PHOTO";
   }
   else if (0 == strncmp (pszColorModeName, JOBKEY_GRAPH, 5))
   {
      pszRet = "GRAPH";
   }
   else if (0 == strncmp (pszColorModeName, JOBKEY_NONE, 4))
   {
      pszRet = "NONE";
   }
///else if (  (  DEVICEID_STYLUS_C60 == pInstance->eDeviceID
///           || DEVICEID_STYLUS_C70 == pInstance->eDeviceID
///           )
///        && 0 == strncmp (pszColorModeName, JOBKEY_AUTO, 4)
///        )
///{
///   pszRet = "AUTO";
///}

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": returning " << pszRet << std::endl;
#endif

   free ((void *)pszColorModeName);

   return pszRet;
}

PSZCRO
convertToEpsonMicroWeave (PSZCRO pszMicroWeaveName)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": high-speed = " << SAFE_PRINT_PSZ(pszMicroWeaveName) << std::endl;
#endif

   PSZRO pszRet = 0;

   if (0 == strncmp (pszMicroWeaveName, JOBKEY_ON, 2))
   {
      pszRet = "ON";
   }
   else if (0 == strncmp (pszMicroWeaveName, JOBKEY_OFF, 3))
   {
      pszRet = "OFF";
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": returning " << pszRet << std::endl;
#endif

   return pszRet;
}

/****************************************************************************/
/* Epson hook points                                                        */
/****************************************************************************/

GModule *
initializeInstance (PSZCRO   pszDeviceName,
                    PHANDLE  phInstance)
{
   GModule        *hModule        = 0;
   PEPSONINSTANCE  pInstance      = 0;
   PSZRO           pszLibraryName = 0;
   DEVICEID        eDeviceID      = DEVICEID_NONE;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << SAFE_PRINT_PSZ(pszDeviceName) << ")" << std::endl;
#endif

   if (0 == strcmp (pszDeviceName, "Epson Stylus Color 980"))
   {
      pszLibraryName = "libsc980.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_980;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Color 880"))
   {
      pszLibraryName = "libsc880.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_880;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Color 777"))
   {
      pszLibraryName = "libsc680_777.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_777;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Color 680"))
   {
      pszLibraryName = "libsc680_777.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_680;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Color 580"))
   {
      pszLibraryName = "libsc580.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_580;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Color 480"))
   {
      pszLibraryName = "libsc480.so";
      eDeviceID      = DEVICEID_STYLUS_COLOR_480;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus C20"))
   {
      pszLibraryName = "libsc20s.so";
      eDeviceID      = DEVICEID_STYLUS_C20;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus C40"))
   {
      pszLibraryName = "libsc40s.so";
      eDeviceID      = DEVICEID_STYLUS_C40;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus C60"))
   {
      pszLibraryName = "libsc60s.so";
      eDeviceID      = DEVICEID_STYLUS_C60;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus C70"))
   {
      pszLibraryName = "libsc70s.so";
      eDeviceID      = DEVICEID_STYLUS_C70;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Photo 810"))
   {
      pszLibraryName = "libsp810_820.so";
      eDeviceID      = DEVICEID_STYLUS_PHOTO_810;
   }
   else if (0 == strcmp (pszDeviceName, "Epson Stylus Photo 820"))
   {
      pszLibraryName = "libsp810_820.so";
      eDeviceID      = DEVICEID_STYLUS_PHOTO_820;
   }

   if (!pszLibraryName)
   {
      DebugOutput::getErrorStream () << "Error: Unsupported device name \"" << pszDeviceName << "\"!" << std::endl;
      return 0;
   }

   Omni::openAndTestDeviceLibrary (pszLibraryName, &hModule);

   if (!hModule)
   {
      DebugOutput::getErrorStream () << "Error: Unable to load \"" << pszLibraryName << "\"!" << std::endl;

      return 0;
   }

   pInstance = (PEPSONINSTANCE)malloc (sizeof (EPSONINSTANCE));

   if (pInstance)
   {
      memset (pInstance, 0, sizeof (*pInstance));

      pInstance->eDeviceID = eDeviceID;

      if (hModule)
      {
         g_module_symbol (hModule,
                          "libprt_init",
                          (gpointer *)&pInstance->pfnLibPrt_Init);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_Init = 0x" << std::hex << (int)pInstance->pfnLibPrt_Init << std::dec << std::endl;
#endif

         g_module_symbol (hModule,
                          "libprt_page_init",
                          (gpointer *)&pInstance->pfnLibPrt_Page_Init);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_Page_Init = 0x" << std::hex << (int)pInstance->pfnLibPrt_Page_Init << std::dec << std::endl;
#endif

         g_module_symbol (hModule,
                          "libprt_out",
                          (gpointer *)&pInstance->pfnLibPrt_Out);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_Out = 0x" << std::hex << (int)pInstance->pfnLibPrt_Out << std::dec << std::endl;
#endif

         g_module_symbol (hModule,
                          "libprt_out_at",
                          (gpointer *)&pInstance->pfnLibPrt_Out_At);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_Out_At = 0x" << std::hex << (int)pInstance->pfnLibPrt_Out_At << std::dec << std::endl;
#endif

         g_module_symbol (hModule,
                          "libprt_page_end",
                          (gpointer *)&pInstance->pfnLibPrt_Page_End);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_Page_End = 0x" << std::hex << (int)pInstance->pfnLibPrt_Page_End << std::dec << std::endl;
#endif

         g_module_symbol (hModule,
                          "libprt_end",
                          (gpointer *)&pInstance->pfnLibPrt_End);

#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pfnLibPrt_End = 0x" << std::hex << (int)pInstance->pfnLibPrt_End << std::dec << std::endl;
#endif
      }

      pInstance->pszQuality    = newCopy (vpszDefaultQuality);
      pInstance->pszHighSpeed  = newCopy (vpszDefaultHighSpeed);
      pInstance->pszColorMode  = newCopy (vpszDefaultColorMode);
      pInstance->pszMicroWeave = newCopy (vpszDefaultMicroWeave);

      pInstance->fInAutoMode   = false;
      pInstance->fInMonoMode   = false;

      pInstance->eState        = STATE_START;
   }

   *phInstance = (HANDLE)pInstance;

   return hModule;
}

PSZCRO
getDriverName ()
{
   return "Epson";
}

std::string *
getJobProperties (HANDLE hInstance)
{
   PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ")" << std::endl;
#endif

   if (!pInstance)
   {
      return 0;
   }

   std::ostringstream oss;

   oss << JOBPROP_QUALITY
       << "="
       << pInstance->pszQuality;
   oss << " ";
   oss << JOBPROP_HIGH_SPEED
       << "="
       << pInstance->pszHighSpeed;
   oss << " ";
   oss << JOBPROP_COLOR_MODE
       << "="
       << pInstance->pszColorMode;
   if (DEVICEID_STYLUS_C70 == pInstance->eDeviceID)
   {
      oss << " ";
      oss << JOBPROP_MICRO_WEAVE
          << "="
          << pInstance->pszMicroWeave;
   }

   return new std::string (oss.str ());
}

bool
setJobProperties (HANDLE              hInstance,
                  PSZCRO              pszJobProperties,
                  DeviceObjectHolder *pDOH)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", \"" << SAFE_PRINT_PSZ(pszJobProperties) << "\")" << std::endl;
#endif

   if (hInstance)
   {
      PEPSONINSTANCE    pInstance   = (PEPSONINSTANCE)hInstance;
      DeviceForm       *pForm       = pDOH->getCurrentForm ();
      DeviceMedia      *pMedia      = pDOH->getCurrentMedia ();
      DevicePrintMode  *pPrintMode  = pDOH->getCurrentPrintMode ();
      DeviceResolution *pResolution = pDOH->getCurrentResolution ();
      DeviceTray       *pTray       = pDOH->getCurrentTray ();
      HardCopyCap      *pHCC        = pForm->getHardCopyCap ();
      PSZRO             pszValue    = 0;
      PSZRO             pszResult   = 0;

      pInstance->hMagicCookie = 0;
      if (!pInstance->pfOutput)
      {
         pInstance->pfOutput = fdopen (STDOUT_FILENO, "wb");
      }
      pInstance->fShouldCloseOutput = false;

#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->pfOutput = 0x" << std::hex << (int)pInstance->pfOutput << std::dec << std::endl;
#endif

      pInstance->init.output = pInstance->pfOutput;

      pszValue = getValue (pszJobProperties, "Form");
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting Form from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pInstance->init.paper = pszValue;
      }
      else
      {
         pInstance->init.paper = newCopy (pForm->getDeviceID ());
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper = " << SAFE_PRINT_PSZ(pInstance->init.paper) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, "printmode");
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting printmode from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pInstance->init.format = pszValue;
      }
      else
      {
         pInstance->init.format = newCopy (pPrintMode->getDeviceID ());
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.format = " << SAFE_PRINT_PSZ(pInstance->init.format) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, "Resolution");
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting Resolution from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pInstance->init.resol = pszValue;
      }
      else
      {
         pInstance->init.resol = newCopy (pResolution->getDeviceID ());
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.resol = " << SAFE_PRINT_PSZ(pInstance->init.resol) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, "InputTray");
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting InputTray from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pInstance->init.bin_id = pszValue;
      }
      else
      {
         pInstance->init.bin_id = newCopy (pTray->getDeviceID ());
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.bin_id = " << SAFE_PRINT_PSZ(pInstance->init.bin_id) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, "media");
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting media from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pInstance->init.media_type = pszValue;
      }
      else
      {
         pInstance->init.media_type = newCopy (pMedia->getDeviceID ());
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.media_type = " << SAFE_PRINT_PSZ(pInstance->init.media_type) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, JOBPROP_HIGH_SPEED);
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting high-speed from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pszResult = convertToEpsonHighSpeed (newCopy (pszValue));
      }
      if (pszValue)
      {
         free ((void *)pInstance->pszHighSpeed);
         pInstance->pszHighSpeed    = pszValue;
         pInstance->init.high_speed = pszResult;
      }
      else
      {
         pInstance->init.high_speed = convertToEpsonHighSpeed (newCopy (vpszDefaultHighSpeed));
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.high_speed = " << SAFE_PRINT_PSZ(pInstance->init.high_speed) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, JOBPROP_QUALITY);
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting quality from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pszResult = convertToEpsonQuality (newCopy (pszValue));
      }
      if (pszResult)
      {
         free ((void *)pInstance->pszQuality);
         pInstance->pszQuality  = pszValue;
         pInstance->init.qlevel = pszResult;
      }
      else
      {
         pInstance->init.qlevel = convertToEpsonQuality (newCopy (vpszDefaultQuality));
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.qlevel = " << SAFE_PRINT_PSZ(pInstance->init.qlevel) << std::endl;
#endif
      }

      pszValue = getValue (pszJobProperties, JOBPROP_COLOR_MODE);
#ifndef RETAIL
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting colormode from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
      if (pszValue)
      {
         pszResult = convertToEpsonColorMode (pInstance, newCopy (pszValue));
      }
      if (pszResult)
      {
         free ((void *)pInstance->pszColorMode);
         pInstance->pszColorMode = pszValue;
      }

      pInstance->init.margin.x = 3;
      pInstance->init.margin.y = 3;

      if (0 == strcmp (pInstance->init.paper, "PP200"))
      {
         pInstance->init.margin.y = 14;
      }

      pInstance->init.paper_size.x = (long)(pHCC->getCx () / 25400.0 * pResolution->getXRes ());
      pInstance->init.paper_size.y = (long)(pHCC->getCy () / 25400.0 * pResolution->getYRes ());
      if (0 == strcmp (pInstance->init.bin_id, "T4AUTO"))
      {
         pInstance->init.paper_area.x = pInstance->init.paper_size.x;
         pInstance->init.paper_area.y = pInstance->init.paper_size.y;

         pInstance->init.margin.x = 0;
         pInstance->init.margin.y = 0;
      }
      else
      {
         pInstance->init.paper_area.x = (long)(  ( pHCC->getCx ()
                                                 - pHCC->getLeftClip ()
                                                 - pHCC->getRightClip ()
                                                 )
                                               / 25400.0
                                               * pResolution->getXRes ());
         pInstance->init.paper_area.y = (long)(  ( pHCC->getCy ()
                                                 - pHCC->getTopClip ()
                                                 - pHCC->getBottomClip ()
                                                 )
                                               / 25400.0
                                               * pResolution->getYRes ());

         pInstance->init.margin.x = (int)( (float)pInstance->init.margin.x
                                         / 25.4
                                         * pResolution->getXRes ()
                                         );
         pInstance->init.margin.y = (int)( (float)pInstance->init.margin.y
                                         / 25.4
                                         * pResolution->getYRes ()
                                         );
      }
      pInstance->init.mw_type = 0;

      if (DEVICEID_STYLUS_C70 == pInstance->eDeviceID)
      {
         pszValue = getValue (pszJobProperties, JOBPROP_MICRO_WEAVE);
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": getting micro-weave from jobprops = " << SAFE_PRINT_PSZ(pszValue) << std::endl;
#endif
         if (pszValue)
         {
            free ((void *)pInstance->pszMicroWeave);
            pInstance->pszMicroWeave = newCopy (pszValue);
            pInstance->init.mw_type  = convertToEpsonMicroWeave (pszValue);
         }
         else
         {
            pInstance->init.mw_type = convertToEpsonMicroWeave (newCopy (vpszDefaultMicroWeave));
#ifndef RETAIL
            if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.mw_type = " << SAFE_PRINT_PSZ(pInstance->init.mw_type) << std::endl;
#endif
         }
      }

      pInstance->fInMonoMode = false;
      if (0 == strcmp (pInstance->init.format, "MONO"))
      {
         pInstance->fInMonoMode = true;
      }

      return true;
   }

   return false;
}

class ArrayJPEnumeration : public Enumeration
{
public:
   ArrayJPEnumeration (PSZCRO *aElements,
                       int     cElements)
   {
      iElement_d  = 0;
      aElements_d = aElements;
      cElements_d = cElements;
   }

   virtual ~
   ArrayJPEnumeration ()
   {
   }

   virtual bool
   hasMoreElements ()
   {
      if (iElement_d < cElements_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (iElement_d > cElements_d - 1)
         return 0;

      return (void *)new JobProperties (aElements_d[iElement_d++]);
   }

private:
   int     iElement_d;
   int     cElements_d;
   PSZCRO *aElements_d;
};

Enumeration *
enumerateInstanceProps (HANDLE hInstance,
                        bool   fInDeviceSpecific)
{
   PEPSONINSTANCE  pInstance = (PEPSONINSTANCE)hInstance;
   EnumEnumerator *pRet      = new EnumEnumerator ();

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", fInDeviceSpecific = " << fInDeviceSpecific << ")" << std::endl;
#endif

   pRet->addElement (new ArrayJPEnumeration (apszQualities,  (int)dimof (apszQualities)));
   pRet->addElement (new ArrayJPEnumeration (apszHighSpeeds, (int)dimof (apszHighSpeeds)));
   pRet->addElement (new ArrayJPEnumeration (apszColorModes, (int)dimof (apszColorModes)));

   if (DEVICEID_STYLUS_C70 == pInstance->eDeviceID)
   {
      pRet->addElement (new ArrayJPEnumeration (apszMicroWeaves, (int)dimof (apszMicroWeaves)));
   }

   return pRet;
}

std::string *
getJobPropertyType (HANDLE  hInstance,
                    PSZCRO  pszKey)
{
   PEPSONINSTANCE      pInstance = (PEPSONINSTANCE)hInstance;
   std::string        *pstrRet   = 0;
   std::ostringstream  oss;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", pszKey = " << SAFE_PRINT_PSZ(pszKey) << ")" << std::endl;
#endif

   if (pInstance)
   {
      if (0 == strcasecmp (pszKey, JOBPROP_QUALITY))
      {
         oss << "string ";
         oss << pInstance->pszQuality;
      }
      else if (0 == strcasecmp (pszKey, JOBPROP_HIGH_SPEED))
      {
         oss << "string ";
         oss << pInstance->pszHighSpeed;
      }
      else if (0 == strcasecmp (pszKey, JOBPROP_COLOR_MODE))
      {
         oss << "string ";
         oss << pInstance->pszColorMode;
      }
      else if (  0 == strcasecmp (pszKey, JOBPROP_MICRO_WEAVE)
              && DEVICEID_STYLUS_C70 == pInstance->eDeviceID
              )
      {
         oss << "string ";
         oss << pInstance->pszMicroWeave;
      }
   }

   // Does it have something in it?
   if (oss.str ()[0])
   {
      pstrRet = new std::string (oss.str ());
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << SAFE_PRINT_PSZ(pszKey) << ") = " << SAFE_PRINT_STRING(pstrRet) << std::endl;
#endif

   return pstrRet;
}

PSZCRO
getJobProperty (HANDLE  hInstance,
                PSZCRO  pszKey)
{
   PEPSONINSTANCE  pInstance = (PEPSONINSTANCE)hInstance;
   PSZRO           pszRet    = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", pszKey = " << SAFE_PRINT_PSZ(pszKey) << ")" << std::endl;
#endif

   if (pInstance)
   {
      if (0 == strcasecmp (pszKey, JOBPROP_QUALITY))
      {
         pszRet = pInstance->pszQuality;
      }
      else if (0 == strcasecmp (pszKey, JOBPROP_HIGH_SPEED))
      {
         pszRet = pInstance->pszHighSpeed;
      }
      else if (0 == strcasecmp (pszKey, JOBPROP_COLOR_MODE))
      {
         pszRet = pInstance->pszColorMode;
      }
      else if (  0 == strcasecmp (pszKey, JOBPROP_MICRO_WEAVE)
              && DEVICEID_STYLUS_C70 == pInstance->eDeviceID
              )
      {
         pszRet = pInstance->pszMicroWeave;
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << SAFE_PRINT_PSZ(pszKey) << ") = " << SAFE_PRINT_PSZ(pszRet) << std::endl;
#endif

   return pszRet;
}

std::string *
translateKeyValue (HANDLE          hInstance,
                   StringResource *pSR,
                   PSZCRO          pszKeyValue)
{
   PEPSONINSTANCE  pInstance = (PEPSONINSTANCE)hInstance;
   std::string    *pstrRet   = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", pszKeyValue = " << SAFE_PRINT_PSZ(pszKeyValue) << ")" << std::endl;
#endif

   if (  pInstance
      && pSR
      )
   {
      PSZRO pszKey           = 0;
      PSZRO pszValue         = 0;
      PSZRO pszEquals        = 0;
      PSZRO pszKeyExternal   = 0;
      PSZRO pszValueExternal = 0;
      int   iID              = StringResource::DEVICE_COMMON_UNKNOWN;

      pszEquals = strchr (pszKeyValue, '=');
      if (pszEquals)
      {
         pszKey   = newCopy (pszKeyValue, pszEquals);
         pszValue = newCopy (pszEquals + 1);
      }
      else
      {
         pszKey = newCopy (pszKeyValue);
      }

      if (pszKey)
      {
         iID = StringResource::DEVICE_COMMON_UNKNOWN;
         if (0 == strcasecmp (pszKey, JOBPROP_QUALITY))
         {
            iID = StringResource::DEVICE_COMMON_QUALITY;
         }
         else if (0 == strcasecmp (pszKey, JOBPROP_HIGH_SPEED))
         {
            iID = StringResource::DEVICE_COMMON_HIGH_SPEED;
         }
         else if (0 == strcasecmp (pszKey, JOBPROP_COLOR_MODE))
         {
            iID = StringResource::DEVICE_COMMON_COLOR_MODE;
         }
         else if (  0 == strcasecmp (pszKey, JOBPROP_MICRO_WEAVE)
                 && DEVICEID_STYLUS_C70 == pInstance->eDeviceID
                 )
         {
            iID = StringResource::DEVICE_COMMON_MICRO_WEAVE;
         }
         if (StringResource::DEVICE_COMMON_UNKNOWN != iID)
         {
            pszKeyExternal = pSR->getStringV (StringResource::STRINGGROUP_DEVICE_COMMON,
                                              iID);
         }

         if (pszKeyExternal)
         {
            pstrRet = new std::string (pszKeyExternal);
         }

         free ((void *)pszKey);
      }

      if (pszValue)
      {
         iID = StringResource::DEVICE_COMMON_UNKNOWN;
         if (0 == strcasecmp (pszValue, JOBKEY_DRAFT))
         {
            iID = StringResource::DEVICE_COMMON_DRAFT;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_STANDARD))
         {
            iID = StringResource::DEVICE_COMMON_STANDARD;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_FINE))
         {
            iID = StringResource::DEVICE_COMMON_FINE;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_SUPER_FINE))
         {
            iID = StringResource::DEVICE_COMMON_SUPER_FINE;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_PHOTO))
         {
            iID = StringResource::DEVICE_COMMON_PHOTO;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_SUPER_PHOTO))
         {
            iID = StringResource::DEVICE_COMMON_SUPER_PHOTO;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_ON))
         {
            iID = StringResource::DEVICE_COMMON_ON;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_OFF))
         {
            iID = StringResource::DEVICE_COMMON_OFF;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_AUTO))
         {
            iID = StringResource::DEVICE_COMMON_AUTO;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_PHOTO))
         {
            iID = StringResource::DEVICE_COMMON_PHOTO;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_GRAPH))
         {
            iID = StringResource::DEVICE_COMMON_GRAPH;
         }
         else if (0 == strcasecmp (pszValue, JOBKEY_NONE))
         {
            iID = StringResource::DEVICE_COMMON_NONE;
         }
         if (StringResource::DEVICE_COMMON_UNKNOWN != iID)
         {
            pszValueExternal = pSR->getStringV (StringResource::STRINGGROUP_DEVICE_COMMON,
                                                iID);
         }

         if (pszValueExternal)
         {
            if (pstrRet)
            {
               *pstrRet += "=";
               *pstrRet += pszValueExternal;
            }
            else
            {
               pstrRet = new std::string (pszValueExternal);
            }
         }

         free ((void *)pszValue);
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << SAFE_PRINT_PSZ(pszKeyValue) << ") = " << SAFE_PRINT_STRING(pstrRet) << std::endl;
#endif

   return pstrRet;
}

bool
initDevice (HANDLE hInstance)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ")" << std::endl;
#endif

   if (hInstance)
   {
      PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;
      int            rc;

      rc = pInstance->pfnLibPrt_Init (&pInstance->hMagicCookie, &pInstance->init);

#ifndef RETAIL
      if (  DebugOutput::shouldOutputPDCBlitter ()
         && rc
         )
      {
         DebugOutput::getErrorStream () << "Error: pfnLibPrt_Init rc = " << rc << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper        = " << SAFE_PRINT_PSZ(pInstance->init.paper)        << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.format       = " << SAFE_PRINT_PSZ(pInstance->init.format)       << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.resol        = " << SAFE_PRINT_PSZ(pInstance->init.resol)        << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.high_speed   = " << SAFE_PRINT_PSZ(pInstance->init.high_speed)   << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.bin_id       = " << SAFE_PRINT_PSZ(pInstance->init.bin_id)       << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.media_type   = " << SAFE_PRINT_PSZ(pInstance->init.media_type)   << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.qlevel       = " << SAFE_PRINT_PSZ(pInstance->init.qlevel)       << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.margin.x     = " << pInstance->init.margin.x                     << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.margin.y     = " << pInstance->init.margin.y                     << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper_size.x = " << pInstance->init.paper_size.x                 << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper_size.y = " << pInstance->init.paper_size.y                 << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper_area.x = " << pInstance->init.paper_area.x                 << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.paper_area.y = " << pInstance->init.paper_area.y                 << std::endl;
         DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->init.mw_type      = " << SAFE_PRINT_PSZ(pInstance->init.mw_type)      << std::endl;
      }
      if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": pInstance->hMagicCookie = 0x" << std::hex << (int)pInstance->hMagicCookie << std::dec << std::endl;
#endif

      if (rc)
      {
         DebugOutput::getErrorStream () << "Warning: Job property combination for InitDevice is not valid!" << std::endl;
      }

      pInstance->eState = STATE_INITED;

      return rc == 0;
   }

   return false;
}

bool
startPage (HANDLE hInstance)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ")" << std::endl;
#endif

   if (hInstance)
   {
      PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;

      if (STATE_START == pInstance->eState)
      {
         if (!initDevice (hInstance))
         {
            return false;
         }
      }

      if (  pInstance->pfnLibPrt_Page_Init
         && pInstance->hMagicCookie
         )
      {
         LIBPRT_PAGE page;
         int         rc;

         page.color_mode    = convertToEpsonColorMode (pInstance, newCopy (pInstance->pszColorMode));
         page.halftone_type = "ED";   // @TBD
         page.brightness    = 0;      // @TBD
         page.contrast      = 0;      // @TBD
         page.saturation    = 0;      // @TBD
         page.r_strength    = 0;      // @TBD
         page.g_strength    = 0;      // @TBD
         page.b_strength    = 0;      // @TBD
         page.reserved      = 0;

         rc = pInstance->pfnLibPrt_Page_Init (pInstance->hMagicCookie, &page);

#ifndef RETAIL
         if (  DebugOutput::shouldOutputPDCBlitter ()
            && rc
            )
         {
            DebugOutput::getErrorStream () << "Error: pfnLibPrt_Page_Init rc = " << rc << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.color_mode    = " << page.color_mode    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.halftone_type = " << page.halftone_type << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.brightness    = " << page.brightness    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.contrast      = " << page.contrast      << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.saturation    = " << page.saturation    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.r_strength    = " << page.r_strength    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.g_strength    = " << page.g_strength    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.b_strength    = " << page.b_strength    << std::endl;
            DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": page.reserved      = " << page.reserved      << std::endl;
         }
#endif

         if (rc)
         {
            DebugOutput::getErrorStream () << "Error: Job property combination for StartPage is not valid!" << std::endl;
         }

         if (0 == strcmp (page.color_mode, "AUTO"))
         {
            pInstance->fInAutoMode = true;
         }

         if (rc)
            return false;
         else
            return true;
      }
   }

   return false;
}

bool
endPage (HANDLE hInstance,
         int    iCanceled)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", " << iCanceled << ")" << std::endl;
#endif

   if (hInstance)
   {
      PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;

      if (  pInstance->pfnLibPrt_Page_End
         && pInstance->hMagicCookie
         )
      {
         int rc;

         rc = pInstance->pfnLibPrt_Page_End (pInstance->hMagicCookie, iCanceled);

#ifndef RETAIL
         if (rc) if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "Error: pfnLibPrt_Page_End rc = " << rc << std::endl;
#endif

         return true;
      }
   }

   return false;
}

bool
beginJob (HANDLE hInstance)
{
   return true;
}

bool
endJob (HANDLE hInstance)
{
   return true;
}

bool
abortJob (HANDLE hInstance)
{
   return endPage (hInstance, 1);
}

int
rasterize (HANDLE       hInstance,
           PBYTE        pbBits,
           PBITMAPINFO2 pbmi2,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ())
      DebugOutput::getErrorStream () << std::hex
           << "OmniEpsonBlitter::" << __FUNCTION__ << " (0x"
           << (int)hInstance << ", 0x"
           << (int)pbBits << ", {"
           << std::dec << pbmi2->cx << ", "
           << pbmi2->cy << ", "
           << pbmi2->cPlanes << ", "
           << pbmi2->cBitCount << "}, "
           << "{" << prectlPageLocation->xLeft << ", " << prectlPageLocation->yBottom << ", " << prectlPageLocation->xRight << ", " << prectlPageLocation->yTop << "})"
           << std::endl;
#endif

   if (!hInstance)
      return false;

   PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;
   PFNLIBPRT_OUT  pfnPrtOut = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ":fInAutoMode = " << pInstance->fInAutoMode << ", fInMonoMode = " << pInstance->fInMonoMode << std::endl;
#endif

   if (pInstance->fInAutoMode)
   {
      pfnPrtOut = pInstance->pfnLibPrt_Out_At;
   }
   else
   {
      pfnPrtOut = pInstance->pfnLibPrt_Out;
   }

   if (  !pfnPrtOut
      || !pInstance->hMagicCookie
      )
      return false;

   int         iSrcWidth;
   LIBPRT_BAND band;
#ifndef RETAIL
   int         rc;
#endif

   iSrcWidth = ((pbmi2->cBitCount * pbmi2->cx + 31) >> 5) << 2;

   if (pInstance->fInMonoMode)
   {
      pbmi2->cBitCount = 8;
   }

   band.width_pixel   = pbmi2->cx;
   band.width_bytes   = ((pbmi2->cBitCount * pbmi2->cx + 31) >> 5) << 2;;
   band.line_number   = 1;
   band.byte_data     = 0;
   band.lp_atbmp_band = 0;

   if (pInstance->fInAutoMode)
   {
      pInstance->bandAt.width_bytes = ((pbmi2->cx * 8 + 31) >> 5) << 2;

      if (2 * pInstance->bandAt.width_bytes > pInstance->cbAllocatedBandAt)
      {
         if (pInstance->bandAt.byte_data)
         {
            free ((void *)pInstance->bandAt.byte_data);
            pInstance->bandAt.byte_data = 0;
         }

         pInstance->cbAllocatedBandAt = 2 * pInstance->bandAt.width_bytes;

         pInstance->bandAt.byte_data = (unsigned char *)calloc (1, pInstance->cbAllocatedBandAt);
      }

      band.lp_atbmp_band = &pInstance->bandAt;
   }

   for (int y = pbmi2->cy - 1; y >= 0; y--)
   {
      band.byte_data = pbBits + y * iSrcWidth;

      if (pInstance->fInMonoMode)
      {
         // convert 24 bit BGR to 8 bit gray scale
         typedef struct _RGB {
            byte b;
            byte g;
            byte r;
         } __attribute__ ((aligned (1)))
           __attribute__ ((packed))
           RGB, *PRGB;

         PRGB  pRGBIn = (PRGB)band.byte_data;
         byte *pKOut  = band.byte_data;

         for (int x = 0; x < pbmi2->cx; x++)
         {
            *pKOut = (byte)( 0.299 * (float)pRGBIn->r
                           + 0.587 * (float)pRGBIn->g
                           + 0.114 * (float)pRGBIn->b
                           );
            pKOut++;
            pRGBIn++;
         }
      }
      else
      {
         // convert 24 bit BGR to 24 bit RGB
         for (int x = 0; x < pbmi2->cx; x++)
         {
            byte *pRGB = band.byte_data + x * 3;
            byte  b    = *pRGB;

            *pRGB       = *(pRGB + 2);
            *(pRGB + 2) = b;
         }
      }

#ifndef RETAIL
//////if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": band.width_pixel = " << band.width_pixel << std::endl;
//////if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": band.width_bytes = " << band.width_bytes << std::endl;
//////if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": band.line_number = " << band.line_number << std::endl;
//////if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": band.byte_data   = 0x" << std::hex << (int)band.byte_data << std::dec << std::endl;
#endif

#ifndef RETAIL
      rc =
#endif
           pfnPrtOut (pInstance->hMagicCookie, &band);

#ifndef RETAIL
      if (rc) if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "Error: pfnPrtOut rc = " << rc << std::endl;
#endif
   }

   return true;
}

bool
freeInstance (GModule *hmodDevice,
              HANDLE   hInstance)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hmodDevice << ", " << hInstance << ")" << std::dec << std::endl;
#endif

   if (hInstance)
   {
      PEPSONINSTANCE pInstance = (PEPSONINSTANCE)hInstance;

      if (  pInstance->pfnLibPrt_End
         && pInstance->hMagicCookie
         )
      {
         int rc;

         rc = pInstance->pfnLibPrt_End (pInstance->hMagicCookie);

#ifndef RETAIL
         if (rc) if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "Error: pfnLibPrt_End rc = " << rc << std::endl;
#endif
      }

      if (  pInstance->pfOutput
         && pInstance->fShouldCloseOutput
         )
      {
         fclose (pInstance->pfOutput);
      }

      if (pInstance->pszQuality)
      {
         free ((void *)pInstance->pszQuality);
         pInstance->pszQuality = 0;
      }
      if (pInstance->pszHighSpeed)
      {
         free ((void *)pInstance->pszHighSpeed);
         pInstance->pszHighSpeed = 0;
      }
      if (pInstance->pszColorMode)
      {
         free ((void *)pInstance->pszColorMode);
         pInstance->pszColorMode = 0;
      }
      if (pInstance->pszMicroWeave)
      {
         free ((void *)pInstance->pszMicroWeave);
         pInstance->pszMicroWeave = 0;
      }

      if (pInstance->bandAt.byte_data)
      {
         free ((void *)pInstance->bandAt.byte_data);
         pInstance->bandAt.byte_data = 0;
      }

      if (pInstance->init.paper)
      {
         free ((void *)pInstance->init.paper);
         pInstance->init.paper = 0;
      }
      if (pInstance->init.format)
      {
         free ((void *)pInstance->init.format);
         pInstance->init.format = 0;
      }
      if (pInstance->init.resol)
      {
         free ((void *)pInstance->init.resol);
         pInstance->init.resol = 0;
      }
      if (pInstance->init.bin_id)
      {
         free ((void *)pInstance->init.bin_id);
         pInstance->init.bin_id = 0;
      }
      if (pInstance->init.media_type)
      {
         free ((void *)pInstance->init.media_type);
         pInstance->init.media_type = 0;
      }

      free ((void *)pInstance);
   }

   if (hmodDevice)
   {
      g_module_close (hmodDevice);
   }

   return false;
}

bool
setOutputStream (HANDLE hInstance,
                 int    fdHandle)
{
   PEPSONINSTANCE pInstance   = (PEPSONINSTANCE)hInstance;
   FILE          *fpNewHandle = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << " (" << std::hex << hInstance << std::dec << ", fdHandle = " << fdHandle << ")" << std::endl;
#endif

   if (hInstance)
   {
      fpNewHandle = fdopen (fdHandle, "wb");

      if (fpNewHandle)
      {
         pInstance->pfOutput           = fpNewHandle;
         pInstance->init.output        = fpNewHandle;
         pInstance->fShouldCloseOutput = false;

         return true;
      }
      else
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputPDCBlitter ()) DebugOutput::getErrorStream () << "OmniEpsonBlitter::" << __FUNCTION__ << ": PDCCMD_SET_OUTPUT_STREAM failed!" << std::endl;
#endif
      }
   }

   return false;
}

void
initializePDCBlitter (PPDC_B_HOOKPOINTS pHookPoints)
{
   pHookPoints->pfnInitializeInstance        = initializeInstance;
   pHookPoints->pfnGetDriverName             = getDriverName;
   pHookPoints->pfnGetJobProperties          = getJobProperties;
   pHookPoints->pfnSetJobProperties          = setJobProperties;
   pHookPoints->pfnEnumerateInstanceProps    = enumerateInstanceProps;
   pHookPoints->pfnGetJobPropertyType        = getJobPropertyType;
   pHookPoints->pfnGetJobProperty            = getJobProperty;
   pHookPoints->pfnTranslateKeyValue         = translateKeyValue;
   pHookPoints->pfnStartPage                 = startPage;
   pHookPoints->pfnEndPage                   = endPage;
   pHookPoints->pfnBeginJob                  = beginJob;
   pHookPoints->pfnEndJob                    = endJob;
   pHookPoints->pfnAbortJob                  = abortJob;
   pHookPoints->pfnRasterize                 = rasterize;
   pHookPoints->pfnFreeInstance              = freeInstance;
   pHookPoints->pfnSetOutputStream           = setOutputStream;
}
